import { PolygonUtil } from "../PolygonUtil";
import Utils from "../Utils";
import PolygonTexture from "../图片切割/PolygonTexture";

const { ccclass, property } = cc._decorator;

@ccclass
export default class RandomVecInPolygon extends cc.Component {

    @property(PolygonTexture)
    polygonTexture: PolygonTexture = null;
    @property(cc.Graphics)
    graphics: cc.Graphics = null;
    @property({ type: cc.Node, displayName: '绿点' })
    greenPoint: cc.Node = null;
    @property({ type: cc.Node, displayName: '蓝点' })
    bluePoint: cc.Node = null;

    private textures: PolygonTexture[] = [];


    onLoad() {
        this.textures.push(this.polygonTexture);
    }
    onDestroy() {
    }

    protected start(): void {
        this.randomPos(this.polygonTexture.polygon);
    }

    /**
     * @多边形内随机点
     * @param vecs @要求vecs是有序的点
     * @所谓的有序就是相邻点在数组内是相邻的
     */
    randomPos(vecs: cc.Vec2[]): cc.Vec2 {
        /**@步骤一  @找到边界 */
        /**@最小x和最大x */
        let minX, maxX;
        /**@最小y和最大y */
        let minY, maxY;
        vecs.forEach(v => {
            if (v.x < minX || !minX) {
                minX = v.x;
            }
            if (v.x > maxX || !maxX) {
                maxX = v.x;
            }
            if (v.y < minY || !minY) {
                minY = v.y;
            }
            if (v.y > maxY || !maxY) {
                maxY = v.y;
            }
        })

        /**@步骤二  @随机x坐标 */
        let x = Utils.randomFloat(minX, maxX);

        /**@步骤三 @过x位置做一条垂直的直线 */
        this.graphics.moveTo(x, minY);
        this.graphics.lineTo(x, maxY);
        this.graphics.stroke();


        let result = [];
        let points: cc.Vec2[] = Utils.deepCopy(vecs);
        /**
         * @把第一个点push到最后保证所有两点之间的直线都判断过
         * @把第一个点push进去是为了形成闭环简化代码处理
         */
        points.push(cc.v2(vecs[0].x, vecs[0].y));
        for (let i = 0; i < points.length - 1; i++) {
            let from = points[i];
            let to = points[i + 1];
            /**
             *  @返回值 [n,p] 
             * @n 0相交，1在共有点，-1不相交  
             * @p交点
             * */
            let [n, p] = PolygonUtil.lineCrossPoint(cc.v2(x, minY), cc.v2(x, maxY), from, to);
            if (n != -1) {
                /** @两线段在某线段顶点相交  @为了处理区间问题需要额外多push一个点 */
                if (n == 1) {
                    result.push(p);
                }
                result.push(p);
            }
        }

        /**@蓝点是随机出的某点 */
        this.bluePoint.setPosition(x, 0);
        cc.log(`Blade:交点个数:${result.length},所有交点:${JSON.stringify(result)}`)
        if (result.length >= 2) {
            /**@从小到大排序 */
            result.sort((a, b) => {
                return a.y - b.y;
            })
            /**@偶数个交点区间 @随机出某个区间 */
            let rIndex = Utils.randomInt(0, Math.floor(result.length / 2) - 1);
            /**@从区间内随机出具体的y坐标 */
            let y = Utils.randomFloat(result[rIndex * 2].y, result[rIndex * 2 + 1].y);
            /**@随机出的目标点 */
            this.greenPoint.setPosition(x, y);
            cc.log(`Blade:交点坐标:(${x},${y})`)
            return cc.v2(x, y);
        }
        else {
            console.error(`出错了`)
        }
    }

}
